@@ -50,7 +50,12 @@ using namespace Konsole;
 #define loc(X, Y) ((Y)*_columns + (X))
 #endif
 
-const Character Screen::DefaultChar =
+const Character Screen::DefaultChar = Character(' ',
+                                                CharacterColor(COLOR_SPACE_DEFAULT, DEFAULT_FORE_COLOR),
+                                                CharacterColor(COLOR_SPACE_DEFAULT, DEFAULT_BACK_COLOR),
+                                                DEFAULT_RENDITION | RE_TRANSPARENT,
+                                                0);
+const Character Screen::VisibleChar =
     Character(' ', CharacterColor(COLOR_SPACE_DEFAULT, DEFAULT_FORE_COLOR), CharacterColor(COLOR_SPACE_DEFAULT, DEFAULT_BACK_COLOR), DEFAULT_RENDITION, 0);
 
 Screen::Screen(int lines, int columns)
@@ -62,6 +67,7 @@ Screen::Screen(int lines, int columns)
     , _scrolledLines(0)
     , _lastScrolledRegion(QRect())
     , _droppedLines(0)
+    , _fastDroppedLines(0)
     , _oldTotalLines(0)
     , _isResize(false)
     , _enableReflowLines(false)
@@ -1338,9 +1344,14 @@ int Screen::droppedLines() const
 {
     return _droppedLines;
 }
+int Screen::fastDroppedLines() const
+{
+    return _fastDroppedLines;
+}
 void Screen::resetDroppedLines()
 {
     _droppedLines = 0;
+    _fastDroppedLines = 0;
 }
 void Screen::resetScrolledLines()
 {
@@ -1489,7 +1500,7 @@ void Screen::clearImage(int loca, int loce, char c, bool resetLineRendition)
 
     // if the character being used to clear the area is the same as the
     // default character, the affected _lines can simply be shrunk.
-    const bool isDefaultCh = (clearCh == Screen::DefaultChar);
+    const bool isDefaultCh = (clearCh == Screen::DefaultChar || clearCh == Screen::VisibleChar);
 
     for (int y = topLine; y <= bottomLine; ++y) {
         const int endCol = (y == bottomLine) ? loce % _columns : _columns - 1;
@@ -1889,6 +1900,7 @@ Character Screen::getCharacter(int col, int row) const
             _history->getCells(row, col, 1, &ch);
         } else {
             ch = Character();
+            ch.rendition.f.transparent = 1;
         }
     }
     return ch;
@@ -2090,6 +2102,78 @@ Character *Screen::getCharacterBuffer(const int size)
     return characterBuffer.data();
 }
 
+QList<int> Screen::getCharacterCounts() const
+{
+    QList<int> counts;
+    int totalLines = _history->getLines() + getLines();
+
+    for (int line = 0; line < totalLines; ++line) {
+        int count = getLineLength(line);
+        bool lineIsWrapped = false;
+        Character *characterBuffer = getCharacterBuffer(count - 1);
+
+        Q_ASSERT(count >= 0);
+
+        if (line < _history->getLines()) {
+            // safety checks
+            Q_ASSERT(count <= _history->getLineLen(line));
+
+            _history->getCells(line, 0, count, characterBuffer);
+
+            // Exclude trailing empty cells from count and don't bother processing them further.
+            // See the comment on the similar case for screen lines for an explanation.
+            while (count > 0 && (characterBuffer[count - 1].flags & EF_REAL) == 0) {
+                count--;
+            }
+
+            if (_history->isWrappedLine(line)) {
+                lineIsWrapped = true;
+            }
+        } else {
+            int screenLine = line - _history->getLines();
+
+            Q_ASSERT(screenLine <= _screenLinesSize);
+
+            screenLine = qMin(screenLine, _screenLinesSize);
+
+            auto *data = _screenLines[screenLine].data();
+            int length = _screenLines.at(screenLine).count();
+
+            // Exclude trailing empty cells from count and don't bother processing them further.
+            // This is necessary because a newline gets added to the last line when
+            // the selection extends beyond the last character (last non-whitespace
+            // character when TrimTrailingWhitespace is true), so the returned
+            // count from this function must not include empty cells beyond that
+            // last character.
+            while (length > 0 && (data[length - 1].flags & EF_REAL) == 0) {
+                length--;
+            }
+
+            if (_lineProperties.at(screenLine).flags.f.wrapped == 1) {
+                lineIsWrapped = true;
+            }
+
+            // count cannot be any greater than length
+            count = qBound(0, length, count);
+        }
+
+        // If the last character is wide, account for it
+        if (Character::width(characterBuffer[count - 1].character, _ignoreWcWidth) == 2)
+            count++;
+
+        // When users ask not to preserve the linebreaks, they usually mean:
+        // `treat LINEBREAK as SPACE, thus joining multiple _lines into
+        // single line in the same way as 'J' does in VIM.`
+        if (_blockSelectionMode || !lineIsWrapped) {
+            ++count;
+        }
+
+        counts.append(count);
+    }
+
+    return counts;
+}
+
 int Screen::copyLineToStream(int line,
                              int start,
                              int count,
@@ -2261,8 +2345,12 @@ void Screen::fastAddHistLine()
 
     // If _history size > max history size it will drop a line from _history.
     // We need to verify if we need to remove a URL.
-    if (removeLine && _escapeSequenceUrlExtractor) {
-        _escapeSequenceUrlExtractor->historyLinesRemoved(1);
+    if (removeLine) {
+        if (_escapeSequenceUrlExtractor) {
+            _escapeSequenceUrlExtractor->historyLinesRemoved(1);
+        }
+
+        _fastDroppedLines++;
     }
     // Rotate left + clear the last line
     std::rotate(_screenLines.begin(), _screenLines.begin() + 1, _screenLines.end());
@@ -2475,6 +2563,7 @@ void Screen::addPlacement(QPixmap pixmap,
                           int &cols,
                           int row,
                           int col,
+                          enum TerminalGraphicsPlacement_t::source source,
                           bool scrolling,
                           int moveCursor,
                           bool leaveText,
@@ -2516,6 +2605,7 @@ void Screen::addPlacement(QPixmap pixmap,
     p->scrolling = scrolling;
     p->X = X;
     p->Y = Y;
+    p->source = source;
 
     if (!leaveText) {
         eraseBlock(row, col, rows, cols);
@@ -2523,7 +2613,13 @@ void Screen::addPlacement(QPixmap pixmap,
     addPlacement(p);
     int needScroll = qBound(0, row + rows - _lines, rows);
     if (moveCursor && scrolling && needScroll > 0) {
-        scrollUp(needScroll);
+        while (needScroll > 0) {
+            scrollUp(qMin(needScroll, _lines));
+            if (!leaveText) {
+                eraseBlock(qMax(0, _lines - needScroll - 1), col, needScroll + 1, cols);
+            }
+            needScroll -= _lines;
+        }
     }
     if (moveCursor) {
         if (rows - needScroll - 1 > 0) {
